@@ -73,9 +73,9 @@ time_t      last_disconnect = 0;
 #endif
 
 typedef struct {
-	GIOChannel *source;
-	GList      *subscriptions;
-	guint       ref_count;
+	MappingProtocolChannel *source;
+	GList                  *subscriptions;
+	guint                   ref_count;
 } MappingConnection;
 
 typedef struct {
@@ -221,6 +221,14 @@ subscription_get_path (MappingSubscription *sub)
 	return sub->path;
 }
 
+static void *
+subscription_get_userdata (MappingSubscription *sub)
+{
+	g_return_val_if_fail (sub != NULL, NULL);
+
+	return sub->userdata;
+}
+
 static gboolean
 subscription_send_event (MappingSubscription *sub,
 			 const char          *filename,
@@ -987,6 +995,29 @@ connection_remove_subscription (MappingConnection   *connection,
 	subscription_unref (sub);
 }
 
+static GnomeVFSResult
+remove_subscription (MappingConnection   *connection,
+		     MappingSubscription *sub)
+{
+	VirtualRoot *root;
+	VirtualNode *node;
+
+	root = lookup_root (sub->rootname, TRUE);
+
+	node = virtual_node_lookup (root, sub->path, NULL);
+
+	subscription_cancel (sub);
+
+	if (node != NULL) {
+		virtual_node_remove_subscription (node, sub);
+	}
+
+	connection_remove_subscription (connection, sub);
+	subscription_unref (sub);
+
+	return GNOME_VFS_OK;
+}
+
 static void
 connection_free (MappingConnection *connection)
 {
@@ -998,12 +1029,17 @@ connection_free (MappingConnection *connection)
 
 	/* release subscriptions that weren't released by cancelling the monitor */
 	while ((cur = g_list_first (connection->subscriptions)) != NULL) {
+		GnomeVFSResult       res;
 		MappingSubscription *sub = cur->data;
 
-		monitor_cancel (sub->rootname, sub->path, connection, NULL);
+		DEBUG_PRINT ("Removing subscription %p\n", sub);
+		res = remove_subscription (connection, sub);
+		if (GNOME_VFS_OK != res) {
+			DEBUG_PRINT ("Removing connection failed: %d\n", res);
+		}
 	}
 
-	g_io_channel_unref (connection->source);
+	mapping_protocol_channel_unref (connection->source);
 
 	g_free (connection);
 	connection = NULL;
@@ -1031,7 +1067,7 @@ connection_unref (MappingConnection *connection)
 }
 
 static MappingConnection *
-connection_new (GIOChannel *source)
+connection_new (MappingProtocolChannel *source)
 {
 	MappingConnection *connection;
 
@@ -1040,7 +1076,8 @@ connection_new (GIOChannel *source)
 	connection = g_new0 (MappingConnection, 1);
 	connection->ref_count = 1;
 	connection->source = source;
-	g_io_channel_ref (source);
+
+	mapping_protocol_channel_ref (source);
 
 	DEBUG_PRINT ("New connection %p\n", connection);
 	return connection;
@@ -1048,7 +1085,8 @@ connection_new (GIOChannel *source)
 
 static MappingSubscription *
 connection_get_subscription (MappingConnection *connection,
-			     const char        *path)
+			     const char        *path,
+			     void              *userdata)
 {
 	GList *l;
 
@@ -1057,8 +1095,10 @@ connection_get_subscription (MappingConnection *connection,
 	for (l = connection->subscriptions; l; l = l->next) {
 		MappingSubscription *sub = l->data;
 
-		if (strcmp (subscription_get_path (sub), path) == 0)
+		if ((strcmp (subscription_get_path (sub), path) == 0)
+		    && (subscription_get_userdata (sub) == userdata)) {
 			return sub;
+		}
 	}
 
 	return NULL;
@@ -1133,31 +1173,20 @@ monitor_cancel (const char        *rootname,
 		MappingConnection *connection,
 		void              *userdata)
 {
-	VirtualRoot         *root;
-	VirtualNode         *node;
 	MappingSubscription *sub;
+	GnomeVFSResult       res;
 
 	DEBUG_PRINT ("Cancelling monitor for %s userdata %p\n", path, userdata);
 
-	root = lookup_root (rootname, TRUE);
+	sub = connection_get_subscription (connection, path, userdata);
 
-	node = virtual_node_lookup (root, path, NULL);
-
-	if (node == NULL) {
+	if (!sub) {
 		return GNOME_VFS_ERROR_NOT_FOUND;
 	}
 
-	sub = connection_get_subscription (connection, path);
+	res = remove_subscription (connection, sub);
 
-	if (!sub)
-		return GNOME_VFS_ERROR_NOT_FOUND;
-
-	subscription_cancel (sub);
-	virtual_node_remove_subscription (node, sub);
-	connection_remove_subscription (connection, sub);
-	subscription_unref (sub);
-
-	return GNOME_VFS_OK;
+	return res;
 }
 
 static void
@@ -1227,7 +1256,7 @@ handle_request (GIOChannel   *source,
 		return FALSE;
 	}
 
-	res = mapping_protocol_request_decode (source, &req);
+	res = mapping_protocol_request_decode (connection->source, &req);
 
 	if (res != 0) {
 		connection_error (connection, condition, NULL);
@@ -1300,7 +1329,7 @@ handle_request (GIOChannel   *source,
 	DEBUG_PRINT ("Encoding reply to request op %s path1: %s result: %d path: '%s' n_strings: %d\n",
 		     operation_string (req.operation), req.path1, reply.result, reply.path, reply.n_strings);
 
-	res = mapping_protocol_reply_encode (source, &reply);
+	res = mapping_protocol_reply_encode (connection->source, &reply);
 
 	mapping_protocol_request_destroy (&req);
 	mapping_protocol_reply_destroy (&reply);
@@ -1321,10 +1350,10 @@ handle_new_client (GIOChannel   *source,
 		   GIOCondition  condition,
 		   gpointer      data)
 {
-	MappingConnection *connection;
-	GIOChannel        *io;
-	int                master_fd = g_io_channel_unix_get_fd (source);
-	int                client;
+	MappingConnection      *connection;
+	MappingProtocolChannel *io;
+	int                     master_fd = g_io_channel_unix_get_fd (source);
+	int                     client;
 
 	/* wait for a client to talk to us */
         if ((client = accept (master_fd, 0, 0)) == -1) {
@@ -1332,13 +1361,14 @@ handle_new_client (GIOChannel   *source,
 		return TRUE;
         }
 
-	io = g_io_channel_unix_new (client);
+	io = mapping_protocol_channel_new (client);
 	connection = connection_new (io);
-	g_io_add_watch (io, G_IO_IN | G_IO_ERR | G_IO_HUP, handle_request, connection);
+
+	g_io_add_watch (io->iochannel, G_IO_IN | G_IO_ERR | G_IO_HUP, handle_request, connection);
 
 	DEBUG_PRINT ("Handling new client source: %p\n", io);
 
-	g_io_channel_unref (io);
+	mapping_protocol_channel_unref (io);
 
 	connection_add (connection);
 	
@@ -1412,7 +1442,7 @@ int
 main (int   argc,
       char *argv[])
 {
-        struct sockaddr_un sin;
+	struct sockaddr_un sin;
 	int                master_socket;
 	GIOChannel        *master_io;
 	GMainLoop         *loop;
@@ -1430,9 +1460,9 @@ main (int   argc,
 #endif
 	
 	if ((master_socket = socket (AF_UNIX, SOCK_STREAM, 0)) == -1) {
-                perror ("Master socket creation failed");
-                exit (1);
-        }
+		perror ("Master socket creation failed");
+		exit (1);
+	}
 	
 	path = mapping_protocol_get_unix_name ();
 
@@ -1442,13 +1472,16 @@ main (int   argc,
 
 	g_unlink (sin.sun_path);
 	if (bind (master_socket, (const struct sockaddr *)&sin, sizeof (sin)) == -1) {
-                perror ("Failed to bind master socket");
-                exit (1);
-        }
+		perror ("Failed to bind master socket");
+		close (master_socket);
+		exit (1);
+	}
+
 	if (listen (master_socket, 5) == -1) {
 		perror ("Failed to listen to master socket");
-                exit (1);
-        }
+		close (master_socket);
+		exit (1);
+	}
 
 	/* Trigger launching app */
 	if (pipe_fd >= 0) {
