@@ -47,7 +47,7 @@ mapping_protocol_get_unix_name (void)
 
 /* adapted from gamin_data_available() */
 int
-mapping_protocol_data_available (GIOChannel *channel)
+mapping_protocol_data_available (MappingProtocolChannel *channel)
 {
 	fd_set         read_set;
 	struct timeval tv;
@@ -56,7 +56,12 @@ mapping_protocol_data_available (GIOChannel *channel)
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	fd = g_io_channel_unix_get_fd (channel);
+	/* First check our read buffers */
+	if (channel->read_len > 0) {
+		return 1;
+	}
+
+	fd = g_io_channel_unix_get_fd (channel->iochannel);
 
  retry:
 	FD_ZERO (&read_set);
@@ -76,10 +81,59 @@ mapping_protocol_data_available (GIOChannel *channel)
 	return (1);
 }
 
+MappingProtocolChannel *
+mapping_protocol_channel_new (int fd)
+{
+	MappingProtocolChannel *channel;
+	GIOChannel             *iochannel = NULL;
+
+	channel = g_new0 (MappingProtocolChannel, 1);
+
+	iochannel = g_io_channel_unix_new (fd);
+	g_io_channel_set_close_on_unref (iochannel, TRUE);
+
+	channel->iochannel = iochannel;
+	channel->read_ptr = NULL;
+	channel->read_len = 0;
+	channel->ref_count = 1;
+
+	return channel;
+}
+
+static void
+mapping_protocol_channel_free (MappingProtocolChannel *channel)
+{
+	g_io_channel_unref (channel->iochannel);
+
+	g_free (channel);
+	channel = NULL;
+}
+
+void
+mapping_protocol_channel_ref (MappingProtocolChannel *channel)
+{
+	g_return_if_fail (channel != NULL);
+	g_return_if_fail (channel->ref_count > 0);
+
+	channel->ref_count += 1;
+}
+
+void
+mapping_protocol_channel_unref (MappingProtocolChannel *channel)
+{
+	g_return_if_fail (channel != NULL);
+	g_return_if_fail (channel->ref_count > 0);
+
+	if (channel->ref_count > 1)
+		channel->ref_count -= 1;
+	else
+		mapping_protocol_channel_free (channel);
+}
+
 static int
-write_all (GIOChannel *channel,
-	   char	      *buf,
-	   int         len)
+write_all (MappingProtocolChannel *channel,
+	   char	                  *buf,
+	   int                     len)
 {
 	int bytes;
 	int bytes_written;
@@ -87,7 +141,7 @@ write_all (GIOChannel *channel,
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	fd = g_io_channel_unix_get_fd (channel);
+	fd = g_io_channel_unix_get_fd (channel->iochannel);
 
 	bytes = 0;
 	while (bytes < len) {
@@ -103,34 +157,49 @@ write_all (GIOChannel *channel,
 }
 
 static int
-read_all (GIOChannel *channel,
-	  char	     *buf,
-	  int         len)
+read_all (MappingProtocolChannel *channel,
+	  char                   *buf,
+	  int                     len)
 {
-	int bytes;
 	int bytes_read;
 	int fd;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	fd = g_io_channel_unix_get_fd (channel);
+	while (len) {
+		if (channel->read_len) {
+			const int tocopy = MIN (len, channel->read_len);
 
-	bytes = 0;
-	while (bytes < len) {
-		bytes_read = read (fd, buf + bytes, len - bytes);
+			memcpy (buf, channel->read_ptr, tocopy);
+			channel->read_len -= tocopy;
+			channel->read_ptr += tocopy;
+			buf += tocopy;
+			len -= tocopy;
+			
+			if (len == 0)
+				break;
+		}
+
+		/* Still more to read in, so re-fill the buffer */
+		fd = g_io_channel_unix_get_fd (channel->iochannel);
+
+		bytes_read = read (fd, channel->read_buffer, MAPPING_PROTOCOL_BUFFER_LEN);
 		if (bytes_read <= 0) {
 			g_warning ("Read error: bytes %d: %s", bytes_read, strerror (errno));
 			return -1;
 		}
-		bytes += bytes_read;
+
+		/* Reset the counters */
+		channel->read_ptr = channel->read_buffer;
+		channel->read_len = bytes_read;
 	}
 
 	return 0;
 }
 
 static int
-encode_int (GIOChannel *channel,
-	    gint32      val)
+encode_int (GString *str,
+	    gint32   val)
 {
 	char  *ptr;
 	gint32 val2;
@@ -138,12 +207,13 @@ encode_int (GIOChannel *channel,
 	val2 = g_htonl (val);
 
 	ptr = (char *) &val2;
-	return write_all (channel, ptr, 4);
+	g_string_append_len (str, ptr, 4);
+	return 0;
 }
 
 static int
-decode_int (GIOChannel *channel,
-	    gint32     *out_val)
+decode_int (MappingProtocolChannel *channel,
+	    gint32                 *out_val)
 {
 	unsigned char ptr [4];
 	guint32       val;
@@ -163,27 +233,27 @@ decode_int (GIOChannel *channel,
 }
 
 static int
-encode_string (GIOChannel *channel,
-	       char       *str)
+encode_string (GString *gstring,
+	       char    *str)
 {
 	int len;
 	int res;
 	
 	if (str == NULL) {
-		res = encode_int (channel, -1);
+		res = encode_int (gstring, -1);
 	} else {
 		len = strlen (str);
-		res = encode_int (channel, len);
+		res = encode_int (gstring, len);
 		if (res == 0)
-			res = write_all (channel, str, len);
+			g_string_append (gstring, str);
 	}
 
 	return res;
 }
 
 static int
-decode_string (GIOChannel *channel,
-	       char      **out)
+decode_string (MappingProtocolChannel *channel,
+	       char                  **out)
 {
 	gint32   len;
 	int      res;
@@ -211,8 +281,8 @@ decode_string (GIOChannel *channel,
 }
 
 static int
-encode_pointer (GIOChannel *channel,
-		gpointer    val)
+encode_pointer (GString *str,
+		gpointer val)
 {
 	int len;
 	int res;
@@ -220,20 +290,20 @@ encode_pointer (GIOChannel *channel,
 	/* We use only local socket so no need for network byte order conversion */
 	
 	if (val == NULL) {
-		res = encode_int (channel, -1);
+		res = encode_int (str, -1);
 	} else {
 		len = sizeof (gpointer);
-		res = encode_int (channel, len);
+		res = encode_int (str, len);
 		if (res == 0) {
-			res = write_all (channel, (char *)&val, len);
+			g_string_append_len (str, (char *)&val, len);
 		}
 	}
 	return res;
 }
 
 static int
-decode_pointer (GIOChannel *channel,
-		gpointer   *out_val)
+decode_pointer (MappingProtocolChannel *channel,
+		gpointer               *out_val)
 {
 	int       len;
 	int       res;
@@ -271,51 +341,58 @@ decode_pointer (GIOChannel *channel,
 }
 
 int
-mapping_protocol_request_encode (GIOChannel *channel,
-				 gint32      operation,
-				 char       *root,
-				 char       *path1,
-				 char       *path2,
-				 gboolean    option,
-				 void       *userdata)
+mapping_protocol_request_encode (MappingProtocolChannel *channel,
+				 gint32                  operation,
+				 char                   *root,
+				 char                   *path1,
+				 char                   *path2,
+				 gboolean                option,
+				 void                   *userdata)
 {
-	int res;
+	int      res;
+	GString *str;
 
 	g_return_val_if_fail (channel != NULL, -1);
+	
+	str = g_string_new (NULL);
 
-	res = encode_int (channel, operation);
+	res = encode_int (str, operation);
 	if (res != 0) return res;
 	
-	res = encode_string (channel, root);
+	res = encode_string (str, root);
 	if (res != 0) return res;
 
-	res = encode_string (channel, path1);
+	res = encode_string (str, path1);
 	if (res != 0) return res;
 
-	res = encode_string (channel, path2);
+	res = encode_string (str, path2);
 	if (res != 0) return res;
 
-	res = encode_int (channel, option);
+	res = encode_int (str, option);
 	if (res != 0) return res;
 
-	res = encode_pointer (channel, userdata);
+	res = encode_pointer (str, userdata);
 	if (res != 0) return res;
 
-        g_io_channel_flush (channel, NULL);
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
+	if (res != 0) return res;
+
+        g_io_channel_flush (channel->iochannel, NULL);
 
 	return 0;
 }
 
 int
-mapping_protocol_request_decode (GIOChannel             *channel,
-				 MappingProtocolRequest *req)
+mapping_protocol_request_decode (MappingProtocolChannel *channel,
+				 MappingProtocolRequest     *req)
 {
 	int res;
 
 	g_return_val_if_fail (channel != NULL, -1);
 	
 	memset (req, 0, sizeof (MappingProtocolRequest));
-	
+
 	res = decode_int (channel, &req->operation);
 	if (res != 0) return res;
 	
@@ -349,22 +426,20 @@ mapping_protocol_request_destroy (MappingProtocolRequest *req)
    there is a listener out there.
  */
 static int
-encode_handshake (GIOChannel *channel,
-		  char        c)
+encode_handshake (GString *str,
+		  char     c)
 {
-	int res;
+	g_string_append_c (str, c);
 
-	res = write_all (channel, &c, 1);
-
-	return res;
+	return 0;
 }
 
 /* this isn't a real handshake but a more of a sanity check to see if
    there is a listener out there.
  */
 static int
-decode_handshake (GIOChannel *channel,
-		  char        c)
+decode_handshake (MappingProtocolChannel *channel,
+		  char                    c)
 {
 	int  res;
 	char buf [2];
@@ -385,35 +460,42 @@ decode_handshake (GIOChannel *channel,
 }
 
 int
-mapping_protocol_reply_encode (GIOChannel           *channel,
-			       MappingProtocolReply *reply)
+mapping_protocol_reply_encode (MappingProtocolChannel *channel,
+			       MappingProtocolReply   *reply)
 {
 	int res;
 	int i;
+	GString * str;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	res = encode_handshake (channel, 'R');
+	str = g_string_new (NULL);
+
+	res = encode_handshake (str, 'R');
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->result);
+	res = encode_int (str, reply->result);
 	if (res != 0) return res;
 	
-	res = encode_string (channel, reply->path);
+	res = encode_string (str, reply->path);
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->option);
+	res = encode_int (str, reply->option);
 	if (res != 0) return res;
 
-	res = encode_int (channel, reply->n_strings);
+	res = encode_int (str, reply->n_strings);
 	if (res != 0) return res;
 
 	for (i = 0; i < reply->n_strings; i++) {
-		res = encode_string (channel, reply->strings [i]);
+		res = encode_string (str, reply->strings [i]);
 		if (res != 0) return res;
 	}
 
-        g_io_channel_flush (channel, NULL);
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
+	if (res != 0) return res;
+
+        g_io_channel_flush (channel->iochannel, NULL);
 	
 	return 0;
 }
@@ -434,8 +516,8 @@ decode_reply_error (MappingProtocolReply *reply,
 }
 
 int
-mapping_protocol_reply_decode (GIOChannel           *channel,
-			       MappingProtocolReply *reply)
+mapping_protocol_reply_decode (MappingProtocolChannel *channel,
+			       MappingProtocolReply       *reply)
 {
 	int i;
 	int res;
@@ -482,7 +564,9 @@ mapping_protocol_reply_decode (GIOChannel           *channel,
 void
 mapping_protocol_reply_destroy (MappingProtocolReply *reply)
 {
+#if 0
 	int i;
+#endif
 
 	g_return_if_fail (reply != NULL);
 
@@ -497,32 +581,39 @@ mapping_protocol_reply_destroy (MappingProtocolReply *reply)
 }
 
 int
-mapping_protocol_monitor_event_encode (GIOChannel                  *channel,
+mapping_protocol_monitor_event_encode (MappingProtocolChannel      *channel,
 				       MappingProtocolMonitorEvent *event)
 {
 	int res;
+	GString *str;
 
 	g_return_val_if_fail (channel != NULL, -1);
 
-	res = encode_handshake (channel, 'E');
+	str = g_string_new (NULL);
+
+	res = encode_handshake (str, 'E');
 	if (res != 0) return res;
 
-	res = encode_int (channel, event->type);
+	res = encode_int (str, event->type);
 	if (res != 0) return res;
 
-	res = encode_pointer (channel, event->userdata);
+	res = encode_pointer (str, event->userdata);
 	if (res != 0) return res;
 
-	res = encode_string (channel, event->path);
+	res = encode_string (str, event->path);
 	if (res != 0) return res;
 
-        g_io_channel_flush (channel, NULL);
+	res = write_all (channel, str->str, str->len);
+	g_string_free (str, TRUE);
+	if (res != 0) return res;
+
+        g_io_channel_flush (channel->iochannel, NULL);
 
 	return 0;
 }
 
 int
-mapping_protocol_monitor_event_decode (GIOChannel                  *channel,
+mapping_protocol_monitor_event_decode (MappingProtocolChannel      *channel,
 				       MappingProtocolMonitorEvent *event)
 {
 	int res;
