@@ -73,6 +73,7 @@ struct _NautilusBurnProcess {
 
 	time_t                    start_time;
 	gint64                    start_num;
+	GList                    *rates;
 
 	gboolean                  changed_text;
 	gboolean                  expect_process_to_die;
@@ -170,15 +171,21 @@ nautilus_burn_process_new ()
 static void
 nautilus_burn_process_free (NautilusBurnProcess *process)
 {
-	if (! process)
+	if (! process) {
 		return;
+	}
 
 	g_free (process->last_error);
 
-	if (process->line)
+	if (process->line) {
 		g_string_free (process->line, TRUE);
-	if (process->line_stderr)
+	}
+
+	if (process->line_stderr) {
 		g_string_free (process->line_stderr, TRUE);
+	}
+
+	g_list_free (process->rates);
 
 	g_free (process->funcs);
 	g_free (process);
@@ -329,8 +336,9 @@ compute_time_remaining (gint64 bytes,
 {
 	long secs;
 
-	if (bytes_per_sec <= 1)
+	if (bytes_per_sec <= 1) {
 		return -1;
+	}
 
 	secs = bytes / bytes_per_sec;
 
@@ -413,6 +421,34 @@ growisofs_blank_stdout_line (NautilusBurnProcess   *process,
 	return TRUE;
 }
 
+static gdouble
+get_average_rate (GList  **rates,
+		  gdouble  rate)
+{
+	const unsigned int max_num = 16;
+	const unsigned int scale  = 1000;
+	gdouble            average;
+	gint32             int_rate;
+	GList             *l;
+
+	if (g_list_length (*rates) > max_num) {
+		*rates = g_list_delete_link (*rates, *rates);
+	}
+	int_rate = (gint32)ceil (scale * rate);
+
+	*rates = g_list_append (*rates, GINT_TO_POINTER (int_rate));
+
+	average = 0;
+	for (l = *rates; l != NULL; l = l->next) {
+		gdouble r = (gdouble)GPOINTER_TO_INT (l->data) / scale;
+		average += r;
+	}
+
+	average /= g_list_length (*rates);
+
+	return average;
+}
+
 static gboolean
 cdrecord_stdout_line (NautilusBurnProcess   *process,
 		      const char            *line,
@@ -437,10 +473,11 @@ cdrecord_stdout_line (NautilusBurnProcess   *process,
 				
 		process->result = NAUTILUS_BURN_RECORDER_RESULT_ERROR;
 				
-		if (recorder->priv->tracks_total_bytes > 0)
+		if (recorder->priv->tracks_total_bytes > 0) {
 			total = recorder->priv->tracks_total_bytes;
-		else
+		} else {
 			total = mb_total * 1048576;
+		}
 
 		if (track > recorder->priv->current_track_num) {
 			recorder->priv->current_track_num = track;
@@ -450,7 +487,16 @@ cdrecord_stdout_line (NautilusBurnProcess   *process,
 		this_remain = (mb_total - mb_written) * 1048576;
 		bytes = (total - recorder->priv->current_track_end_pos) + this_remain;
 
-		secs = compute_time_remaining (bytes, (double)speed * CDR_SPEED);
+		secs = -1;
+		if (speed > 0) {
+			gdouble ave_rate;
+			gdouble rate;
+
+			rate = (double)speed * CDR_SPEED;
+
+			ave_rate = get_average_rate (&process->rates, rate);
+			secs = compute_time_remaining (bytes, ave_rate);
+		}
 
 		if (!process->changed_text) {
 			g_signal_emit (recorder,
@@ -469,9 +515,10 @@ cdrecord_stdout_line (NautilusBurnProcess   *process,
 		g_signal_emit (recorder,
 			       nautilus_burn_recorder_table_signals [PROGRESS_CHANGED], 0,
 			       percent, secs);
-	} else if (sscanf (line, "Total size:  %d MB ", &mb_total)) {
+	} else if (sscanf (line, "Track %*d: %*s %d MB ", &mb_total)) {
+		/* %s can be either "data" or "audio" */
 		if (mb_total > 0) {
-			recorder->priv->tracks_total_bytes = mb_total * 1048576;
+			recorder->priv->tracks_total_bytes += mb_total * 1048576;
 		}
 	} else if (g_str_has_prefix (line, "Re-load disk and hit <CR>") ||
 		   g_str_has_prefix (line, "send SIGUSR1 to continue")) {
